"""
Contains the FocalPlane and WCS classes.
"""

import azcam
import numpy
from azcam.header import Header, ObjectHeaderMethods


class FocalPlane(ObjectHeaderMethods):
    """
    The FocalPlane class describes a focalplane layout.
    """

    def __init__(self):
        """
        Create an instance of FocalPlane class.
        """

        #: focalplane header object
        self.header = Header("Focalplane")

        #: World Coordinate System object
        self.wcs = WCS(self)

        # only needed for server side
        self.header.set_header("focalplane")

        # fixed values
        # for multiple CCD detectors:
        # 0 - each CCD has its own physical coordinates
        # 1 - combine physical coordinates (physical coordinates
        #    are the same as detector coordinates)
        self.split_physical_coords = 0
        #: number of detectors along X axis
        self.numdet_x: int = 1
        #: number of detectors along Y axis
        self.numdet_y: int = 1
        #: amplifier readout orientation (new)
        self.amp_cfg: list = [0]
        # detector number for each detector
        self.det_number = [1]
        # extension number for each amplifier
        self.ext_number = [1]
        # extension name for each amplifier
        self.ext_name = ["1"]
        # extension number order for each amplifier when creating binary image
        self.jpg_ext = [1]
        # detector position in x direction in pixels
        self.detpos_x = [1]
        # detector position in y direction in pixels
        self.detpos_y = [1]
        # amplifier's positions along X axis
        self.extpos_x = [1]
        # amplifier's positions along Y axis
        self.extpos_y = [1]
        # amplifier's positions in pixels along X axis
        self.amppix1 = [1]
        # amplifier's positions in pixels along Y axis
        self.amppix2 = [1]
        # gap between amplifiers in X. value different than 0
        #    means that there is a gap on the left side of the amplifier.
        # gapx is the sum of all gaps on the left side of the amplifier.
        self.gapx = [0]
        # gap between amplifiers in Y. value different than 0 means
        #    that there is a gap below of the amplifier.
        # gapy is the sum of all gaps below the amplifier.
        self.gapy = [0]

        # number of amplifiers along X axis
        self.numamps_x = 1
        # number of amplifiers along Y axis
        self.numamps_y = 1
        # reference (zeropoint) position X (in pixels)
        self.refpix1 = 1
        # reference (zeropoint) position Y (in pixels)
        self.refpix2 = 1
        # unbinned number of visible pixels per amplifier along X axis
        self.ampvispix_x = 1
        # unbinned number of visible pixels per amplifier along Y axis
        self.ampvispix_y = 1

        # amplifer gain values in e/DN
        self.gains = []
        # amplifer read noise in e
        self.rdnoises = []

        # calculated values
        # number of pixels in image
        self.numpix_image = 0
        # number of bytes per image
        self.numbytes_image = 0
        # number of columns per image
        self.numcols_image = 0
        # number of rows per image
        self.numrows_image = 0
        # number of total amps per image
        self.numamps_image = 1
        # number of detectors
        self.num_detectors = 1

        # number of columns per amplifier
        self.numcols_amp = 1
        # number of row per amplifier
        self.numrows_amp = 1
        # number of serial amps per detector
        self.num_ser_amps_det = 1
        # number of parallel amps per detector
        self.num_par_amps_det = 1
        # number of total amps per detector
        self.num_amps_det = 1

        # first column
        self.first_col = 1
        # first row
        self.first_row = 1
        # last column
        self.last_col = 1
        # last row
        self.last_row = 1
        # column binning
        self.col_bin = 1
        # row binning
        self.row_bin = 1

        # number of visible columns per amplifier
        self.numviscols_amp = 1
        # number of visible rows per amplifier
        self.numvisrows_amp = 1
        # number of visible columns per amplifier
        self.numviscols_image = 1
        # number of visible rows per amplifier
        self.numvisrows_image = 1

        # number of colums in overscan
        self.numcols_overscan = 0
        # number of rows in overscan
        self.numrows_overscan = 0
        # number of colums in underscan
        self.numcols_underscan = 0
        # number of rows in underscan
        self.numrows_underscan = 0
        # underscan skips
        self.xunderscan = 0
        self.yunderscan = 0
        # overscan skips
        self.xoverscan = 0
        self.yoverscan = 0

        # bias location constants
        self.BL_COLUNDER = 1
        self.BL_COLOVER = 2
        self.BL_ROWUNDER = 3
        self.BL_ROWOVER = 4
        self.bias_location = self.BL_COLOVER

        # number of pixels per amplifier
        self.numpix_amp = 1
        # number of pixels per detector
        self.numpix_det = 1

        # shifting
        self.ns_total = 0
        self.ns_predark = 0
        self.ns_underscan = 0
        self.ns_overscan = 0
        self.np_total = 0
        self.np_predark = 0
        self.np_underscan = 0
        self.np_overscan = 0
        self.np_frametransfer = 0

        # set shifting parameters
        self.coltotal = 0
        self.colusct = 0
        self.coluscw = 0
        self.coluscm = 0
        self.coloscw = 0
        self.coloscm = 0
        self.rowtotal = 0
        self.rowusct = 0
        self.rowuscw = 0
        self.rowuscm = 0
        self.rowoscw = 0
        self.rowoscm = 0
        self.framet = 0

    def update_header_keywords(self):
        """
        Update focal plane keywords in header
        """

        self.header.set_keyword("AZCAM-HEAD", "OK", "AzCam Header flag", "str")
        self.header.set_keyword(
            "NUM-DETX", self.numdet_x, "Number of detectors in X", "int"
        )
        self.header.set_keyword(
            "NUM-DETY", self.numdet_y, "Number of detectors in Y", "int"
        )
        self.header.set_keyword(
            "NUM-AMPX", self.numamps_x, "Number of amplifiers in X", "int"
        )
        self.header.set_keyword(
            "NUM-AMPY", self.numamps_y, "Number of amplifiers in Y", "int"
        )
        self.header.set_keyword("REF-PIX1", self.refpix1, "Reference pixel 1", "int")
        self.header.set_keyword("REF-PIX2", self.refpix2, "Reference pixel 2", "int")

        return

    def update_ext_keywords(self):
        """
        Update focal plane keywords for single extension
        """
        self.header.set_keyword("AMP-CFG", self.amp_cfg[0], "Amplifier configuration")
        self.header.set_keyword("DET-NUM", self.det_number[0], "Detector number")
        self.header.set_keyword("EXT-NUM", self.ext_number[0], "extension number")
        self.header.set_keyword("JPG-EXT", self.jpg_ext[0], "Image section")
        self.header.set_keyword("DET-POSX", self.detpos_x[0], "Detector position in X")
        self.header.set_keyword("DET-POSY", self.detpos_y[0], "Detector position in Y")
        self.header.set_keyword("Ext-PosX", self.extpos_x[0], "Amplifier position in X")
        self.header.set_keyword("Ext-PosY", self.extpos_y[0], "Amplifier position in Y")
        self.header.set_keyword(
            "AMP-PIX1", self.amppix1[0], "Amplifier pixel position in X"
        )
        self.header.set_keyword(
            "AMP-PIX2", self.amppix2[0], "Amplifier pixel position in Y"
        )

        return

    def update_header(self):
        """
        Update headers, reading current data.
        """

        self.update_header_keywords()

        return

    def set_format(
        self,
        ns_total: int = -1,
        ns_predark: int = -1,
        ns_underscan: int = -1,
        ns_overscan: int = -1,
        np_total: int = -1,
        np_predark: int = -1,
        np_underscan: int = -1,
        np_overscan: int = -1,
        np_frametransfer: int = -1,
    ):
        """
        Set the detector format.
        Args:
            ns_total: number of visible columns
            ns_predark: number of physical dark underscan columns
            ns_underscan: desired number of desired dark underscan columns
            ns_overscan: number of dark overscan columns
            np_total: number of visible rows
            np_predark: number of physical dark underscan rows
            np_underscan: number of desired dark underscan rows
            np_overscan: number of desired dark overscan rows
            np_frametransfer: rows to frame transfer shift
        """

        ns_total = int(ns_total)
        ns_predark = int(ns_predark)
        ns_underscan = int(ns_underscan)
        ns_overscan = int(ns_overscan)
        np_total = int(np_total)
        np_predark = int(np_predark)
        np_underscan = int(np_underscan)
        np_overscan = int(np_overscan)
        np_frametransfer = int(np_frametransfer)

        # defaults
        if ns_total == -1:
            ns_total = self.ns_total
        if ns_predark == -1:
            ns_predark = self.ns_predark
        if ns_underscan == -1:
            ns_underscan = self.ns_underscan
        if ns_overscan == -1:
            ns_overscan = self.ns_overscan
        if np_total == -1:
            np_total = self.np_total
        if np_predark == -1:
            np_predark = self.np_predark
        if np_underscan == -1:
            np_underscan = self.np_underscan
        if np_overscan == -1:
            np_overscan = self.np_overscan
        if np_frametransfer == -1:
            np_frametransfer = self.np_frametransfer

        # sets the format parameters for a detector
        self.ns_total = int(ns_total)
        self.ns_predark = int(ns_predark)
        self.ns_underscan = int(ns_underscan)
        self.ns_overscan = int(ns_overscan)
        self.np_total = int(np_total)
        self.np_predark = int(np_predark)
        self.np_underscan = int(np_underscan)
        self.np_overscan = int(np_overscan)
        self.np_frametransfer = int(np_frametransfer)

        # set shifting parameters
        self.coltotal = int(ns_total)
        self.colusct = int(ns_predark)
        self.coluscw = int(ns_underscan)
        self.coluscm = 0
        self.coloscw = int(ns_overscan)
        self.coloscm = 0
        self.rowtotal = int(np_total)
        self.rowusct = int(np_predark)
        self.rowuscw = int(np_underscan)
        self.rowuscm = 0
        self.rowoscw = int(np_overscan)
        self.rowoscm = 0
        self.framet = int(np_frametransfer)

        return

    def get_format(self):
        """
        Return a list of current detector format parameters
        """

        return [
            self.ns_total,
            self.ns_predark,
            self.ns_underscan,
            self.ns_overscan,
            self.np_total,
            self.np_predark,
            self.np_underscan,
            self.np_overscan,
            self.np_frametransfer,
        ]

    def set_focalplane(
        self, numdet_x=-1, numdet_y=-1, numamps_x=-1, numamps_y=-1, amp_cfg=[0]
    ):
        """
        Sets focal plane configuration. Use after set_format() and before set_roi().
        This command replaces SetConfiguration.
        Default focalplane values are set here.
        numdet_x defines number of detectors in Column direction.
        numdet_y defines number of detectors in Row direction.
        numamps_x defines number of amplifiers in Column direction.
        numamps_y defines number of amplifiers in Row direction.
        amp_cfg defines each amplifier's orientation (ex: [0,1,2,3]).
        0 - normal
        1 - flip x
        2 - flip y
        3 - flip x and y
        """

        if numdet_x == -1:
            numdet_x = self.numdet_x
        if numdet_y == -1:
            numdet_y = self.numdet_y
        if numamps_x == -1:
            numamps_x = self.numamps_x
        if numamps_y == -1:
            numamps_y = self.numamps_y

        self.numdet_x = numdet_x
        self.numdet_y = numdet_y
        self.numamps_x = numamps_x
        self.numamps_y = numamps_y
        self.amp_cfg = amp_cfg

        # set the keywords in the main header
        self.header.set_keyword(
            "N-DET-X", self.numdet_x, "Number of detectors in X", "int"
        )
        self.header.set_keyword(
            "N-DET-Y", self.numdet_y, "Number of detectors in Y", "int"
        )
        self.header.set_keyword(
            "N-AMPS-X", self.numamps_x, "Number of amplifiers in X", "int"
        )
        self.header.set_keyword(
            "N-AMPS-Y", self.numamps_y, "Number of amplifiers in Y", "int"
        )

        # set number of detectors
        self.num_detectors = self.numdet_x * self.numdet_y

        # set amps per detector
        self.num_ser_amps_det = int(self.numamps_x / self.numdet_x)
        self.num_par_amps_det = int(self.numamps_y / self.numdet_y)
        self.num_amps_det = self.num_ser_amps_det * self.num_par_amps_det

        # set amps in focal plane
        self.numamps_image = self.num_detectors * self.num_amps_det

        # set unbinned number of visible pixels per amplifier along X and Y axis
        self.ampvispix_x = int(self.ns_total / self.numamps_x)
        self.ampvispix_y = int(self.np_total / self.numamps_y)

        # set default values
        self.set_default_values()

        # initialize WCS so array size is correct, values may be manually set later
        self.wcs.initialize()

        return

    def get_focalplane(self):
        """
        Returns the focal plane configuration.
        """

        return [
            self.numdet_x,
            self.numdet_y,
            self.numamps_x,
            self.numamps_y,
            self.amp_cfg,
        ]

    def set_roi(
        self,
        first_col=-1,
        last_col=-1,
        first_row=-1,
        last_row=-1,
        col_bin=-1,
        row_bin=-1,
        roi_num=0,
    ):
        """
        Sets the ROI values for a specified ROI.
        Currently only one ROI (0) is supported.
        These values are for the entire focal plane, not just one detector.
        """

        # set current values
        first_col = int(first_col)
        last_col = int(last_col)
        first_row = int(first_row)
        last_row = int(last_row)
        col_bin = int(col_bin)
        row_bin = int(row_bin)
        roi_num = int(roi_num)

        if first_col == -1:
            first_col = self.first_col
        if last_col == -1:
            last_col = self.last_col
        if first_row == -1:
            first_row = self.first_row
        if last_row == -1:
            last_row = self.last_row
        if col_bin == -1:
            col_bin = self.col_bin
        if row_bin == -1:
            row_bin = self.row_bin

        self.first_col = int(first_col)
        self.last_col = int(last_col)
        self.first_row = int(first_row)
        self.last_row = int(last_row)
        self.col_bin = int(col_bin)
        self.row_bin = int(row_bin)
        self.roi_num = int(roi_num)

        # bad fix for ROI now entire focalplane - change later!
        if self.numdet_x > 1:
            lc = last_col / self.numdet_x
            lr = last_row / self.numdet_y
        else:
            lc = last_col
            lr = last_row

        fc = first_col
        fr = first_row
        # calculate variables for shifting a single AMPLIFIER
        self.xunderscan = int(min(self.colusct / self.col_bin, self.coluscw))
        self.xskip = int(
            min(self.colusct - self.xunderscan * self.col_bin, self.coluscm)
        )
        self.xpreskip = self.colusct - self.xskip - self.xunderscan * self.col_bin
        self.xskip += fc - 1
        self.xdata = (lc - (fc - 1)) / self.col_bin
        self.xdata = max(0, self.xdata)
        self.xdata = int(self.xdata / self.num_ser_amps_det)
        self.xpostskip = self.coltotal / self.numamps_x - (
            (fc - 1) + self.xdata * self.col_bin
        )
        self.xpostskip = int(max(0, self.xpostskip))
        self.xpostskip += self.coloscm
        self.xoverscan = self.coloscw

        self.numcols_amp = self.xunderscan + self.xdata + self.xoverscan
        self.numcols_overscan = self.xoverscan
        self.numviscols_amp = self.xdata
        # self.numviscols_image=self.numviscols_amp*self.num_ser_amps_det*self.numamps_x
        self.numviscols_image = (
            self.numviscols_amp * self.num_ser_amps_det * self.numdet_x
        )

        self.yunderscan = int(min(self.rowusct / self.row_bin, self.rowuscw))
        self.yskip = int(
            min(self.rowusct - self.yunderscan * self.row_bin, self.rowuscm)
        )
        self.ypreskip = self.rowusct - self.yskip - self.yunderscan * self.row_bin
        self.yskip += self.first_row - 1
        self.ydata = (lr - (fr - 1)) / self.row_bin
        self.ydata = max(0, self.ydata)
        self.ydata = int(self.ydata / self.num_par_amps_det)
        self.ypostskip = self.rowtotal / self.numamps_y - (
            (fr - 1) + self.ydata * self.row_bin
        )
        self.ypostskip = int(max(0, self.ypostskip))
        self.ypostskip += self.rowoscm
        self.yoverscan = self.rowoscw

        self.numrows_amp = self.yunderscan + self.ydata + self.yoverscan
        self.numrows_overscan = self.yoverscan
        self.numvisrows_amp = self.ydata
        # self.numvisrows_image=self.numvisrows_amp*self.num_par_amps_det*self.numamps_y
        self.numvisrows_image = (
            self.numvisrows_amp * self.num_par_amps_det * self.numdet_y
        )

        self.numpix_amp = self.numcols_amp * self.numrows_amp

        # totals for a single DETECTOR
        self.numcols_det = self.numcols_amp * self.num_ser_amps_det
        self.numrows_det = self.numrows_amp * self.num_par_amps_det
        self.numpix_det = self.numcols_det * self.numrows_det

        # totals for the IMAGE
        self.numpix_image = self.numpix_det * self.num_detectors
        self.numcols_image = self.numcols_det * self.numdet_x
        self.numrows_image = self.numrows_det * self.numdet_y
        self.numbytes_image = self.numpix_image * 2

        # number of UNBINNED pixels to flush during clear
        self.xflush = (
            self.xpreskip
            + self.xskip
            + self.xpostskip
            + (self.xunderscan + self.xdata + self.xoverscan) * self.col_bin
        )
        self.yflush = (
            self.ypreskip
            + self.yskip
            + self.ypostskip
            + (self.yunderscan + self.ydata + self.yoverscan) * self.row_bin
        )

        self.set_amp_positions()

        return

    def get_roi(self, roi_num=0):
        """
        Returns a list of the ROI parameters for the roi_num specified.
        Currently only one ROI (0) is supported.
        Returned list format is (first_col,last_col,first_row,last_row,col_bin,row_bin).
        """

        return [
            self.first_col,
            self.last_col,
            self.first_row,
            self.last_row,
            self.col_bin,
            self.row_bin,
        ]

    def roi_reset(self):
        """
        Resets detector ROI values to full frame, current binning.
        """

        # get focalplane format
        reply = self.get_format()

        # update controller shifting parameters
        self.set_roi(1, reply[0], 1, reply[4], -1, -1)

        return reply

    def set_amp_positions(self):
        """
        Calculates amplifiers positions including gaps between amplifiers and CCDs
        New: Zareba 23Mar2012
        """
        self.amppix1 = numpy.empty(shape=[self.numamps_image], dtype="<u2")
        self.amppix2 = numpy.empty(shape=[self.numamps_image], dtype="<u2")

        indx = 0
        for flip in self.amp_cfg:
            if flip == 0:
                flip_x = 0
                flip_y = 0
            elif flip == 1:
                flip_x = 1
                flip_y = 0
            elif flip == 2:
                flip_x = 0
                flip_y = 1
            else:
                flip_x = 1
                flip_y = 1

            size_x = self.numviscols_amp * self.col_bin
            size_y = self.numvisrows_amp * self.row_bin

            self.amppix1[indx] = (
                (self.extpos_x[indx] - 1) * size_x
                + flip_x * size_x
                + self.gapx[indx]
                + 1
                - flip_x
            )
            self.amppix2[indx] = (
                (self.extpos_y[indx] - 1) * size_y
                + flip_y * size_y
                + self.gapy[indx]
                + 1
                - flip_y
            )

            indx += 1

        return

    def set_default_values(self):
        """
        Sets default values for focalplane variables
        """

        # set default values for the extension names, 'im1' -> 'imN'
        # use _WCS file to set nonstandard values (focalplane.ext_name = [...])
        self.ext_name = numpy.empty(shape=[self.numamps_image], dtype="S16")
        for ext in range(self.numamps_image):
            self.ext_name[ext] = f"IM{(ext + 1)}"
        self.ext_name = [y.decode() for y in self.ext_name]  # new

        # set default values for the etension numbers, 1 -> N
        # use _WCS.py file to set nonstandard values ( focalplane.ext_number = [...])
        self.ext_number = numpy.empty(shape=[self.numamps_image], dtype="<u2")
        for ext in range(self.numamps_image):
            self.ext_number[ext] = ext + 1

        # set default values for the jpg extension values, 1 -> N
        # use _WCS.py file to set nonstandard values (focalplane.jpg_ext = [...])
        self.jpg_ext = numpy.empty(shape=[self.numamps_image], dtype="<u2")
        for ext in range(self.numamps_image):
            self.jpg_ext[ext] = ext + 1

        # set default values for the gaps, 0
        self.gapx = numpy.empty(shape=[self.numamps_image], dtype="<u2")
        for ext in range(self.numamps_image):
            self.gapx[ext] = 0
        self.gapy = numpy.empty(shape=[self.numamps_image], dtype="<u2")
        for ext in range(self.numamps_image):
            self.gapy[ext] = 0

        # set extension positions
        self.extpos_x = numpy.empty(shape=[self.numamps_image], dtype="<u2")
        for ext in range(self.numamps_image):
            self.extpos_x[ext] = 1
        self.extpos_y = numpy.empty(shape=[self.numamps_image], dtype="<u2")
        for ext in range(self.numamps_image):
            self.extpos_y[ext] = 1

        self.det_number = numpy.empty(shape=[self.numamps_image], dtype="<u2")
        numpy.ndarray.fill(self.det_number, 1)
        self.detpos_x = numpy.empty(shape=[self.numamps_image], dtype="<u2")
        numpy.ndarray.fill(self.detpos_x, 1)
        self.detpos_y = numpy.empty(shape=[self.numamps_image], dtype="<u2")
        numpy.ndarray.fill(self.detpos_y, 1)

        return

    def set_extension_name(self, ext_name):
        for i, ext in enumerate(ext_name):
            self.ext_name[i] = ext

        return

    def set_extension_extnum(self, ext_number):
        for i, ext in enumerate(ext_number):
            self.ext_number[i] = ext

        return

    def set_ref_pixel(self, xy):
        """
        Set the reference pixel.
        xy is [X,Y] in pixels.
        """

        self.refpix1 = xy[0]
        self.refpix2 = xy[1]

        return

    def set_extension_position(self, xy):
        """
        Set the extension position of each amplifier.
        xy is [[X,Y]] in index numbers, starting at [1,1].
        """

        for i, xy_ in enumerate(xy):
            self.extpos_x[i] = xy_[0]
            self.extpos_y[i] = xy_[1]

        return

    def set_jpg_order(self, indices):
        """
        Set JPG image positions.
        """

        self.jpg_ext = indices

        return


class WCS(object):
    def __init__(self, FPobject):
        # ref to parent focalplane object
        self.fpobj = FPobject

        # tool which provides RA and DEC info
        #  usually telecope or instrument
        self.coord_object = "telescope"

        # Equinox of WCS - default value 2000
        self.equinox = 2000
        # WCS dimensionality - default value 2
        self.wcs_dim = 2
        # True if CD matrix is used, False if rot_deg and scale1,2 are used
        self.cd_matrix = 0
        # coordinate type - x axis
        self.ctype1 = "RA---TAN"
        # coordinate type - y axis
        self.ctype2 = "DEC--TAN"

        # RA position of image
        self.ra_deg = ""
        # DEC position of image
        self.dec_deg = ""
        # default RA position
        self.def_ra = "00:00:00.00"
        # default DEC position
        self.def_dec = "00:00:00.00"

        # True if CD matrix is used, False if rot_deg and scale1, 2 are used
        self.cd_matrix = 0
        # translation matrix: CD1_1 - number of pixel per detector pixel
        self.cd_1_1 = 1
        # translation matrix: CD1_2 - rotation
        self.cd_1_2 = 1
        # translation matrix: CD2_1 - rotation
        self.cd_2_1 = 1
        # translation matrix: CD2_2 - number of pixel per detector pixel
        self.cd_2_2 = 1

        # rotation in deg
        self.rot_deg = [0]

        # scale - x axis - deg/pixel
        self.scale1 = [1.0 / 3600.0]
        # scale - y axis - deg/pixel
        self.scale2 = [1.0 / 3600.0]

        # CCD to amplifier transformation matrix x
        self.atm_1_1 = [1]
        # CCD to amplifier transformation matrix y
        self.atm_2_2 = [1]
        # CCD to amplifier transformation vector x
        self.atv1 = [1]
        # CCD to amplifier transformation vector y
        self.atv2 = [1]

        # CCD to image transformation matrix x
        self.ltm_1_1 = [1]
        # CCD to image transformation matrix y
        self.ltm_2_2 = [1]
        # CCD to image transformation vector x
        self.ltv_1 = [1]
        # CCD to image transformation vector y
        self.ltv_2 = [1]

        # CCD to detector transformation matrix x
        self.dtm_1_1 = [1]
        # CCD to detector transformation matrix y
        self.dtm_2_2 = [1]
        # CCD to detector transformation vector x
        self.dtv_1 = [1]
        # CCD to detector transformation vector x
        self.dtv_2 = [1]

    def initialize(self):
        numexts = self.fpobj.numamps_image

        self.atm_1_1 = numpy.empty(shape=[numexts], dtype="<i4")
        self.atm_2_2 = numpy.empty(shape=[numexts], dtype="<i4")
        self.atv1 = numpy.empty(shape=[numexts], dtype="<i4")
        self.atv2 = numpy.empty(shape=[numexts], dtype="<i4")

        self.ltm_1_1 = numpy.empty(shape=[numexts], dtype="<f")
        self.ltm_2_2 = numpy.empty(shape=[numexts], dtype="<f")
        self.ltv_1 = numpy.empty(shape=[numexts], dtype="<f")
        self.ltv_2 = numpy.empty(shape=[numexts], dtype="<f")

        self.dtm_1_1 = numpy.empty(shape=[numexts], dtype="<u2")
        self.dtm_2_2 = numpy.empty(shape=[numexts], dtype="<u2")
        self.dtv_1 = numpy.empty(shape=[numexts], dtype="<u2")
        self.dtv_2 = numpy.empty(shape=[numexts], dtype="<u2")

        # rotation in deg
        self.rot_deg = numpy.empty(shape=[numexts], dtype="<f")
        for ext in range(numexts):
            self.rot_deg[ext] = 0.0

        # scale - x axis - deg/pixel
        self.scale1 = numpy.empty(shape=[numexts], dtype="<f")
        for ext in range(numexts):
            self.scale1[ext] = 1.0 / 3600.0
        # scale - y axis - deg/pixel
        self.scale2 = numpy.empty(shape=[numexts], dtype="<f")
        for ext in range(numexts):
            self.scale2[ext] = 1.0 / 3600.0

        return

    def get_ra_dec(self):
        """
        Get RA and DEC from telescope header.
        They should have been copied from telescope header to this image header.
        """

        if self.coord_object is None:
            return

        try:
            get_ra = azcam.db.tools[self.coord_object].header.get_keyword("RA")
            self.ra_deg = self._ra_to_deg(get_ra[0])
        except Exception:
            self.ra_deg = self._ra_to_deg(self.def_ra)

        try:
            get_dec = azcam.db.tools[self.coord_object].header.get_keyword("DEC")
            self.dec_deg = self._dec_to_deg(get_dec[0])
        except Exception:
            self.dec_deg = self._dec_to_deg(self.def_dec)

        return

    def _dec_to_deg(self, dec):
        """
        Convert DEC into decimal degrees
        """

        sgn = +1.0
        dms = [0, 0, 0]

        sign = dec[0]
        if sign == "-":
            sgn = -1.0
            dec = dec[1:]
        else:
            if sign == "+":
                sgn = +1.0
                dec = dec[1:]

        x = dec.split(":")

        itm = len(x)
        for i in range(itm):
            if i <= 2:
                dms[i] = x[i]

        return (int(dms[0]) + int(dms[1]) / 60.0 + float(dms[2]) / 3600.0) * sgn

    def _ra_to_deg(self, ra):
        """
        Convert RA into decimal degrees
        """

        hms = [0, 0, 0]

        x = ra.split(":")

        itm = len(x)
        for i in range(itm):
            if i <= 2:
                hms[i] = x[i]

        return (int(hms[0]) + int(hms[1]) / 60.0 + float(hms[2]) / 3600.0) * 15



"""
Contains the base TempConMag class.
"""

import azcam
import azcam.exceptions
from azcam.tools.tempcon import TempCon


class TempConMag(TempCon):
    """
    Defines the MAG temperature control class.
    """

    def __init__(self, tool_id="tempcon", description="Mag tempcon"):
        super().__init__(tool_id, description)

        self.num_temp_reads = 3
        self.control_temperature = self.bad_temp_value  # no active control

        return

    def get_temperature(self, temperature_id: int = 0):
        """
        Returns the current detector temperature in Celsius.
        """

        if not self.is_enabled:
            # azcam.exceptions.warning("Tempcon not enabled")
            return -999.9

        if not self.is_initialized:
            # azcam.exceptions.warning("Tempcon not initialized")
            return -999.9

        itemp1 = itemp = 0.0
        for _ in range(self.num_temp_reads):  # multiple reads for average
            status = azcam.db.tools["controller"].magio("gcam_get_temp", 0)
            itemp1 = int(status[1])  # get temp in DN
            itemp = itemp + itemp1
        itemp = itemp / self.num_temp_reads  # make average

        if itemp > 8196:  # convert to Celsius
            itemp = itemp - 16384

        temp = (itemp - 902) * (-26.0 / 1164.0)  # float

        temp = self.apply_corrections(temp, 0)

        # make nice float
        temp = float(int(temp * 1000.0) / 1000.0)

        # use some limits
        if temp > 100.0 or temp < -300.0:
            temp = self.bad_temp_value

        return temp



from dash import html, dcc, callback, Input, Output
import dash_bootstrap_components as dbc

import azcam


def status_card():
    """
    Create status card.
    """
    # status card
    style1 = {"font-style": "italic"}

    row1 = html.Tr(
        [
            html.Td("Image Filename", style=style1),
            html.Td("filename here", id="filename_status", colSpan=3),
        ]
    )
    row2 = html.Tr(
        [
            html.Td("Image Title", style=style1),
            html.Td("title here", id="imagetitle_status", colSpan=3),
        ]
    )
    row3 = html.Tr(
        [
            html.Td("Image Type", style=style1),
            html.Td("type here", id="imagetype_status"),
            html.Td("Exposure Time", style=style1),
            html.Td("ET here", id="et_status"),
        ]
    )
    row4 = html.Tr(
        [
            html.Td("Test Image", style=style1),
            html.Td("test here", id="imagetest_status"),
            html.Td("Mode", style=style1),
            html.Td("mode here", id="mode_status"),
        ]
    )
    row5 = html.Tr(
        [
            html.Td("Temperatures", style=style1),
            html.Td("temps here", id="temps_status"),
            html.Td("Binning", style=style1),
            html.Td("binning here", id="binning_status"),
        ]
    )
    row_messsage = html.Tr(
        [
            html.Td("Message", style=style1),
            html.Td("messages here", id="message_status", colSpan=3),
        ]
    )
    row_ts = html.Tr(
        [
            html.Td("Timestamp", style=style1),
            html.Td("timestamp here", id="timestamp_status", colSpan=3),
        ]
    )
    row_progress = html.Tr(
        [
            html.Td("Progress", style=style1),
            html.Td(
                dbc.Progress(id="progress_status"),
                colSpan=3,
            ),
        ]
    )

    table_body = [
        html.Tbody([row1, row2, row3, row4, row5, row_progress, row_ts, row_messsage])
    ]
    table = dbc.Table(
        table_body,
        bordered=True,
        striped=True,
        # dark=False,
        style={"padding": "2em"},
    )
    # table = dbc.Table(table_header + table_body, bordered=True)

    status_card = dbc.Card(
        [
            dbc.CardHeader(
                "Status for Current Exposure", style={"font-weight": "bold"}
            ),
            dbc.CardBody(
                [
                    html.Div(
                        "",
                        id="estate_status",
                        style={"font-weight": "bold"},
                    ),
                    table,
                ]
            ),
        ]
    )

    # status table update
    @callback(
        Output("filename_status", "children"),
        Output("imagetitle_status", "children"),
        Output("imagetype_status", "children"),
        Output("et_status", "children"),
        Output("imagetest_status", "children"),
        Output("mode_status", "children"),
        Output("temps_status", "children"),
        Output("binning_status", "children"),
        Output("timestamp_status", "children"),
        Output("estate_status", "children"),
        Output("progress_status", "value"),
        Output("message_status", "children"),
        Input("expstatus_interval", "n_intervals"),
    )
    def update_expstatus(n):
        """
        Get exposure status and update fields.
        """

        webdata = azcam.db.tools["exposure"].get_status()

        # the return order must match Output order
        filename = webdata["filename"]
        imagetitle = webdata["imagetitle"]
        imagetype = webdata["imagetype"]
        et = webdata["exposuretime"]
        imagetest = webdata["imagetest"]
        mode = webdata["mode"]

        camtemp = float(webdata["camtemp"])
        dewtemp = float(webdata["dewtemp"])
        temps = f"Camera: {camtemp:.01f} C    Dewar: {dewtemp:.01f} C"

        colbin = webdata["colbin"]
        rowbin = webdata["rowbin"]
        binning = f"RowBin: {rowbin}    ColBin: {colbin}"

        ts = webdata["timestamp"]
        estate = webdata["exposurestate"]
        progress = float(webdata["progressbar"])
        message = webdata["message"]

        return [
            filename,
            imagetitle,
            imagetype,
            et,
            imagetest,
            mode,
            temps,
            binning,
            ts,
            estate,
            progress,
            message,
        ]

    return status_card



"""
*azcam.fits* contains FITS image support functions for azcam.
"""

import os
import time
import typing
import warnings

import azcam
import azcam.utils
import azcam.exceptions
import numpy
import numpy.polynomial.polynomial as poly
from astropy.io import fits as pyfits


def file_exists(filename: str) -> bool:
    """
    Args:
        filename: filename to check if exists.
    Returns:
        True if the file exists.
    """

    fe = azcam.utils.make_image_filename(filename)

    return os.path.exists(fe)


# *******************************************************************************
# image header commands
# *******************************************************************************
def get_section(filename: str, section: str, extension: int = 0) -> list:
    """
    Returns image section pixel numbers from a FITS keyword.

    Args:
        filename: image filename.
        section: section name (like "CCDSEC").
        extension: image extension number where keyword is located.
    Returns:
        list of zero-based pixel numbers defining the
        section as `[first_col,last_col,first_row,last_row]`.
    """

    filename = azcam.utils.make_image_filename(filename)

    # get row limits
    datasec = get_keyword(filename, section, extension)
    datasec = datasec.lstrip("[")
    datasec = datasec.split(":")
    first_col = int(datasec[0]) - 1

    datasec1 = datasec[1].split(",")
    last_col = int(datasec1[0]) - 1
    first_row = int(datasec1[1]) - 1
    last_row = int(datasec[2].rstrip("]")) - 1

    return [first_col, last_col, first_row, last_row]


def get_keyword(filename: str, keyword: str, extension: int = 0) -> typing.Any:
    """
    Return a header keyword value.

    Args:
        filename: image filename.
        keyword: keyword name.
        extension: image extension number.
    Returns:
        the specified keyword value.
    """

    Image = azcam.utils.make_image_filename(filename)

    hdr = pyfits.getheader(Image, extension)
    return hdr[keyword]


def edit_keyword(
    filename: str, keyword: str, value: typing.Any, extension: int = 0
) -> None:
    """
    Edits a header keyword value.

    Args:
        filename: image filename.
        keyword: keyword name.
        value: new value of keyword.
        extension: image extension number.
    """

    Image = azcam.utils.make_image_filename(filename)

    with pyfits.open(Image, mode="update") as hdulist:
        prihdr = hdulist[0].header
        prihdr[keyword] = value

    return


def get_header(filename: str, extension: int = 0) -> object:
    """
    Return an image header.

    Args:
        filename: image filename.
        extension: image extension number.
    Returns:
        the image header as a pyfits header object.
    """

    filename = azcam.utils.make_image_filename(filename)

    return pyfits.getheader(filename, extension)


def add_history(filename: str, history_string: str, extension: int = 0) -> None:
    """
    Adds a HISTORY line containing 'history_string' to the image.
    Used to record actions performed on an image. A 20 character time stamp
    is added as a prefix, and the result is split across up to three cards
    if it is too long to fit in one. Any extra text is truncated.

    Args:
        filename: image filename.
        history_string: string to add as a HISTORY keyword.
        extension: image extension number.
    """

    value = (
        time.strftime("%Y/%m/%d %H:%M:%S ", time.gmtime(time.time())) + history_string
    )
    if len(value) > 70:
        value = value[:70] + "\n" + value[70:]
    if len(value) > 141:
        value = value[:141] + "\n" + value[141:]
    if len(value) > 212:
        value = value[:212]

    # write the keyword
    filename = azcam.utils.make_image_filename(filename)

    hdu = pyfits.open(filename, mode="update")
    hdu[extension].header.add_history(value)
    hdu.close()

    return


def get_history(filename: str, extension: int = 0) -> str:
    """
    Returns the HISTOR header lines.

    Args:
        filename: image filename.
        extension: image extension number.
    Returns:
        string containing all HISTORY lines.
    """

    filename = azcam.utils.make_image_filename(filename)

    hdr = pyfits.getheader(filename, extension)

    history = hdr["HISTORY"]

    return str(history)


# **************************************************************************************************
# Image extensions
# **************************************************************************************************


def get_extensions(filename: str) -> list:
    """
    Returns the number of image extensions and their indices.
    The number of extensions is 0 for a standard FITS file and >0 for MEF.
    The first data extension for an MEF file is 1.

    Args:
        filename: image filename.
    Returns:
        the list `[number_exts, first_ext, last_ext]` containing the number of extensions and the
        indices for the python `range` function to iterate over them.
    """

    filename = azcam.utils.make_image_filename(filename)
    with pyfits.open(filename) as hdulist:
        if "NEXTEND" in hdulist[0].header:
            num_ext = hdulist[0].header["NEXTEND"]
            if num_ext == 0:
                first_ext = 0
            else:
                first_ext = 1
            last_ext = num_ext + 1
        else:
            num_ext = 0
            first_ext = 0
            last_ext = 1

    return [num_ext, first_ext, last_ext]


# **************************************************************************************************
# Image math
# **************************************************************************************************


def arith(
    filename1: str,
    operator: str,
    filename2: str,
    filename3: str = "",
    datatype: str = "uint16",
) -> None:
    """
    Simple image arithmetic of FITS files.

    Args:
        filename1: image filename.
        operator: '+','-','/', or '*'.
        filename2: may be an image filename or a constant.
        filename3: optional, must be an image filename.  If not specified, result to filename1.
        datatype: valid datatype string for resultant data type.
    """

    MakeU16 = 1 if datatype == "uint16" else 0
    header = []  # header for output file

    # open Image1
    filename1 = azcam.utils.make_image_filename(filename1)
    numext1, fext, lext = get_extensions(filename1)
    # im1 = pyfits.open(filename1, lazy_load_hdus=False)  # this is an hdulist
    with pyfits.open(filename1, lazy_load_hdus=False) as im1:
        if numext1 > 0:
            MEF = 1
            header.append(im1[0].header)  # save for output file
            data1 = []  # first data index is 0
            for i in range(1, lext):
                header.append(im1[i].header)  # save for output
                if _is_image_extension(im1, i):  # only use image data
                    data1.append(im1[i].data)
        else:
            MEF = 0
            header = im1[0].header  # save for output file
            data1 = im1[0].data

        # make float - new
        if MEF:
            for i in range(len(data1)):
                if _is_image_extension(im1, i):
                    data1[i] = data1[i].astype("float32")
        else:
            data1 = data1.astype("float32")

    # check if filename2 is actually a number and not a filename
    if not isinstance(filename2, str):
        data2 = float(filename2)
        SCALAR = 1

    # open Image2 and get data
    else:
        SCALAR = 0
        filename2 = azcam.utils.make_image_filename(filename2)
        numext2, fext, lext = get_extensions(filename2)
        with pyfits.open(filename2, lazy_load_hdus=False) as im2:
            if numext1 != numext2:
                im2.close()
                raise azcam.exceptions.AzcamError("unequal FITS image extensions")
            if MEF:
                data2 = []
                for i in range(1, lext):
                    if _is_image_extension(im2, i):
                        data2.append(im2[i].data.astype("float32"))
            else:
                data2 = im2[0].data.astype("float32")

    # operate
    NewFile = 1
    if MEF:
        data3 = []
    if operator == "*":
        if MEF:
            for i in range(len(data1)):
                if SCALAR:
                    data3.append(data1[i] * data2)
                else:
                    data3.append(data1[i] * data2[i])
        else:
            data3 = data1 * data2

    elif operator == "+":
        if MEF:
            for i in range(len(data1)):
                if SCALAR:
                    data3.append(data1[i] + data2)
                else:
                    data3.append(data1[i] + data2[i])
        else:
            data3 = data1 + data2
    elif operator == "-":
        if MEF:
            for i in range(len(data1)):
                if SCALAR:
                    data3.append(data1[i] - data2)
                else:
                    data3.append(data1[i] - data2[i])
        else:
            data3 = data1 - data2
    elif operator == "/":
        if MEF:
            for i in range(len(data1)):
                if SCALAR:
                    data3.append(data1[i] / data2)
                else:
                    data3.append(data1[i] / data2[i])
        else:
            data3 = data1 / data2
    # write result (all data is now float32)
    if filename3 == "":
        filename3 = filename1
        os.remove(filename3)
    filename3 = azcam.utils.make_image_filename(filename3)

    if NewFile:
        if MEF:
            phdu = pyfits.PrimaryHDU(None, header[0])
            hdulist = pyfits.HDUList()
            hdulist.append(phdu)
            for i in range(numext1):
                if MakeU16:
                    numpy.clip(data3[i], 0, 2**16, data3[i])  # clip values below zero
                    hdu = pyfits.ImageHDU(data3[i].astype("uint16"), header[i + 1])
                else:
                    hdu = pyfits.ImageHDU(data3[i].astype(datatype), header[i + 1])
                hdulist.append(hdu)
            if len(header) > len(data3):
                for i in range(len(data3) + 1, len(im1)):
                    hdulist.append(im1[i])
            hdulist.writeto(filename3, overwrite=1)
        else:
            im3 = pyfits.PrimaryHDU(data3, header)
            im3.writeto(filename3, overwrite=1)

    return


def _is_image_extension(hdulist: object, extension: int) -> bool:
    """
    Check is an extension of a image hdulist contains image data.

    Args:
        hdulist: image hdulist.
        extension: image extension number.
    Returns:
        True if the image extension is of type IMAGE.
    """

    return (
        "XTENSION" in hdulist[extension].header
        and hdulist[extension].header["XTENSION"] == "IMAGE"
    )


def add(
    filename1: str, filename2: str, filename3: str, datatype: str = "uint16"
) -> None:
    """
    Add two images.
    `filename3 = filename1 + filename2`.

    Args:
        filename1: image filename.
        filename2: may be an image filename or a constant.
        filename3: optional, must be an image filename.  If not specified, result to filename1.
        datatype: valid datatype string for resultant data type.
    """

    return arith(filename1, "+", filename2, filename3, datatype)


def sub(
    filename1: str, filename2: str, filename3: str, datatype: str = "uint16"
) -> None:
    """
    Subtract two images.
    `filename3 = filename1 - filename2`.

    Args:
        filename1: image filename.
        filename2: may be an image filename or a constant.
        filename3: optional, must be an image filename.  If not specified, result to filename1.
        datatype: valid datatype string for resultant data type.
    """

    return arith(filename1, "-", filename2, filename3, datatype)


def mult(
    filename1: str, filename2: str, filename3: str, datatype: str = "uint16"
) -> None:
    """
    Multiple two images.
    `filename3 = filename1 * filename2`.

    Args:
        filename1: image filename.
        filename2: may be an image filename or a constant.
        filename3: optional, must be an image filename.  If not specified, result to filename1.
        datatype: valid datatype string for resultant data type.
    """

    return arith(filename1, "*", filename2, filename3, datatype)


def div(
    filename1: str, filename2: str, filename3: str, datatype: str = "uint16"
) -> None:
    """
    Divide two images.
    `filename3 = filename1 / filename2`.

    Args:
        filename1: image filename.
        filename2: may be an image filename or a constant.
        filename3: optional, must be an image filename.  If not specified, result to filename1.
        datatype: valid datatype string for resultant data type.
    """

    return arith(filename1, "/", filename2, filename3, datatype)


def combine(
    file_list: list = [],
    out_filename: str = "combined.fits",
    combination_type: str = "median",
    overscan_correct: int = 1,
    fit_order=3,
    datatype="float32",
) -> None:
    """
    Make a combination iamge from a list of FITS filenames.

    Args:
        file_list: list of filenames to combine.
        out_filename: output filename.
        combination_type: combination type, "median", "sum", or "mean".
        overscan_correct: line fit order if >0 for overscan correction before combination.
    """

    numfiles = len(file_list)
    if numfiles < 2:
        raise azcam.exceptions.AzcamError("two or more images are required")

    header = []  # header for output file
    dataset = []  # combined data

    for fnum, f in enumerate(file_list):
        filename = azcam.utils.make_image_filename(f)

        # overscan correct
        if overscan_correct > 0:
            colbias(filename, fit_order)

        dataarray = []  # each dataset by channel for each file

        # open each image and get data
        with pyfits.open(filename) as im:
            numexts, firstext, lastext = get_extensions(filename)

            # setup based on first image
            if fnum == 0:
                if numexts > 0:
                    MEF = 1
                    header.append(im[0].header)
                    for i in range(firstext, lastext):
                        header.append(im[i].header)
                        dataarray.append(im[i].data)

                else:
                    MEF = 0
                    header = im[0].header  # save for output file
                    dataarray = im[0].data

            else:
                if numexts > 0:
                    for i in range(numexts):
                        dataarray.append(im[i + 1].data)
                else:
                    # dataarray=dataarray+im.data
                    dataarray = im[0].data

            dataset.append(dataarray)

    if MEF:
        data3d = []
        for chan in range(numexts):
            data3d.append(numpy.array([x[chan] for x in dataset]))
        data_combined = []
        for chan in range(numexts):
            if combination_type == "median":
                data_combined.append(numpy.median(data3d[chan], axis=0))
            elif combination_type == "sum":
                data_combined.append(numpy.sum(data3d[chan], axis=0))
            elif combination_type == "mean":
                data_combined.append(numpy.mean(data3d[chan], axis=0))
    else:
        data3d = numpy.array([x for x in dataset])
        if combination_type == "median":
            data_combined = numpy.median(data3d, axis=0)
        elif combination_type == "sum":
            data_combined = numpy.sum(data3d, axis=0)
        elif combination_type == "mean":
            data_combined = numpy.mean(data3d, axis=0)

    if MEF:
        newdata = []
        for i in range(numexts):
            newdata.append(data_combined[i].astype(datatype))
    else:
        newdata = data_combined.astype(datatype)

    # set output datatype
    # if datatype != "float64":
    #     newdata = [x.astype(datatype) for x in newdata]

    # write result
    if MEF:
        phdu = pyfits.PrimaryHDU(None, header[0])
        hdulist = pyfits.HDUList()
        hdulist.append(phdu)
        for i in range(firstext, lastext):
            hdu = pyfits.ImageHDU(newdata[i - 1], header[i])
            hdulist.append(hdu)

        with warnings.catch_warnings():  # surpress warning
            warnings.simplefilter("ignore")
            hdulist.writeto(out_filename, overwrite=1)
        hdulist.close()
    else:
        im1 = pyfits.PrimaryHDU(newdata, header)
        im1.writeto(out_filename, overwrite=1)

    # update header
    add_history(
        out_filename,
        "COMBINED Data was %s combined from %d images" % (combination_type, numfiles),
        0,
    )

    return


# *********************************************************************************************
# pixel computations - uses ROI
# *********************************************************************************************
def mean(filename: str = "test", roi: list = []) -> list:
    """
    Compute mean of an image ROI in every extension.

    Args:
        filename: image filename.
        roi: Region-Of-Interest.
    Returns:
        list of the means of each image extension or ROI in each extension.
    """

    roi = _get_data_roi(roi)

    means = []
    filename = azcam.utils.make_image_filename(filename)

    with pyfits.open(filename) as im:
        NumExt, first_ext, last_ext = get_extensions(filename)
        for chan in range(first_ext, last_ext):
            data = im[chan].data
            numpy.seterr(under="ignore")
            mean1 = data[roi[0] : roi[1], roi[2] : roi[3]].mean()
            means.append(mean1)

    return means


def sdev(filename: str = "test", roi: list = []) -> list:
    """
    Compute standard deviation of an image ROI in every extension.

    Args:
        filename: image filename.
        roi: Region-Of-Interest.
    Returns:
        list of the standard deviations of each image extension or ROI in each extension.
    """

    roi = _get_data_roi(roi)

    filename = azcam.utils.make_image_filename(filename)

    with pyfits.open(filename) as im:
        sdevs = []
        NumExt, first_ext, last_ext = get_extensions(filename)
        for chan in range(first_ext, last_ext):
            data = im[chan].data
            sdev1 = data[roi[0] : roi[1], roi[2] : roi[3]].std()
            sdevs.append(sdev1)

    return sdevs


def stat(filename: str = "test", roi: list = []) -> list:
    """
    Compute mean and sdev image statistics of ROI in every extension.

    Args:
        filename: image filename.
        roi: Region-Of-Interest.
    Returns:
        list of `[[means], [sdevs], ROI]` for each image extension or ROI in each extension.
    """

    filename = azcam.utils.make_image_filename(filename)
    mean1 = mean(filename, roi)
    sdev1 = sdev(filename, roi)

    return [mean1, sdev1, roi]


def minimum(filename: str = "test", roi: str = []) -> list:
    """
    Compute minimum of an image ROI for every extension.

    Args:
        filename: image filename.
        roi: Region-Of-Interest.
    Returns:
        list of the minima of each image extension or ROI in each extension.
    """

    mins = []

    roi = _get_data_roi(roi)

    filename = azcam.utils.make_image_filename(filename)

    with pyfits.open(filename) as im:
        NumExt, first_ext, last_ext = get_extensions(filename)
        for chan in range(first_ext, last_ext):
            data = im[chan].data
            numpy.seterr(under="ignore")
            dmin = data[roi[0] : roi[1], roi[2] : roi[3]].min()
            mins.append(dmin)

    return mins


def maximum(filename: str = "test", roi: str = []) -> list:
    """
    Compute maximum of an image ROI for every extension.

    Args:
        filename: image filename.
        roi: Region-Of-Interest.
    Returns:
        list of the maxima of each image extension or ROI in each extension.
    """

    maxs = []

    roi = _get_data_roi(roi)

    filename = azcam.utils.make_image_filename(filename)

    with pyfits.open(filename) as im:
        NumExt, first_ext, last_ext = get_extensions(filename)
        for chan in range(first_ext, last_ext):
            data = im[chan].data
            numpy.seterr(under="ignore")
            dmax = data[roi[0] : roi[1], roi[2] : roi[3]].max()
            maxs.append(dmax)

    return maxs


def get_data(filename: str = "test", roi: str = []) -> list:
    """
    Return data (pixel values) from an ROI in an image for every extension.
    NOT FINISHED!

    Args:
        filename: image filename.
        roi: Region-Of-Interest.
    Returns:
        list of pixel values.
    """

    filename = azcam.utils.make_image_filename(filename)
    roi = _get_data_roi(roi)

    with pyfits.open(filename) as im:
        NumExt, first_ext, last_ext = get_extensions(filename)
        if NumExt == 0:
            datalist = im[0].data[roi[0] : roi[1], roi[2] : roi[3]]
        else:
            datalist = [
                im[chan].data[roi[0] : roi[1], roi[2] : roi[3]]
                for chan in range(first_ext, last_ext)
            ]

    return datalist


def resample(filename: str, resample: int = 2) -> None:
    """
    Resample an image by combining adjacent pixels.

    Args:
        filename: image filename.
        resample: number of pixels to combine in each dimension.
    """

    filename = azcam.utils.make_image_filename(filename)
    numext, first_ext, last_ext = get_extensions(filename)

    with pyfits.open(
        filename, "update", memmap=False
    ) as hdul:  # memap solves lock issue
        for chan in range(first_ext, last_ext):
            data = hdul[chan].data
            dims = [hdul[chan].header["NAXIS1"], hdul[chan].header["NAXIS2"]]
            d2 = data.reshape(dims[1], dims[0])

            d2_binned = _bin_ndarray(
                d2, (int(dims[1] / resample), int(dims[0] / resample)), "sum"
            )

            hdul[chan].header["NAXIS1"] = int(dims[0] / resample)
            hdul[chan].header["NAXIS2"] = int(dims[1] / resample)
            hdul[chan].data = d2_binned

    return


def _bin_ndarray(ndarray, new_shape, operation="sum"):
    """
    Bins an ndarray in all axes based on the target shape, by summing or
        averaging.

    Number of output dimensions must match number of input dimensions and
        new axes must divide old ones.

    https://stackoverflow.com/questions/8090229/resize-with-averaging-or-rebin-a-numpy-2d-array

    Example
    -------
    >>> m = np.arange(0,100,1).reshape((10,10))
    >>> n = bin_ndarray(m, new_shape=(5,5), operation='sum')
    >>> print(n)

    [[ 22  30  38  46  54]
     [102 110 118 126 134]
     [182 190 198 206 214]
     [262 270 278 286 294]
     [342 350 358 366 374]]

    """
    operation = operation.lower()
    if operation not in ["sum", "mean"]:
        raise ValueError("Operation not supported.")
    if ndarray.ndim != len(new_shape):
        raise ValueError("Shape mismatch: {} -> {}".format(ndarray.shape, new_shape))
    compression_pairs = [(d, c // d) for d, c in zip(new_shape, ndarray.shape)]
    flattened = [ll for p in compression_pairs for ll in p]
    ndarray = ndarray.reshape(flattened)
    for i in range(len(new_shape)):
        op = getattr(ndarray, operation)
        ndarray = op(-1 * (i + 1))
    return ndarray


def colbias(filename: str = "test", fit_order: int = 3, margin_cols: int = 0) -> None:
    """
    Remove column bias from a FITS file.

    Args:
        filename: image filename.
        fit_order: polynomial fit order, use 0 to remove median not fitted value.
        margin_cols: number of overscan columns to skip before correction.
    """

    filename = azcam.utils.make_image_filename(filename)

    # open image and get data
    with pyfits.open(filename, mode="update") as im:
        # if already COLBIAS then exit here
        try:
            history = im[0].header["History"]
            for h in history:
                h = repr(h)
                if "COLBIAS" in h:
                    return
                else:
                    pass
        except KeyError:
            pass

        numexts, firstext, lastext = get_extensions(filename)

        # find column median value
        if numexts == 0:
            Median = []
        else:
            Median = [[]]  # skip phdu

        # get overscan info
        reply = get_keyword(filename, "OVRSCAN2", firstext)
        if isinstance(reply, str):
            overscanrows = 0
        else:
            overscanrows = int(reply)
        col1, col2, row1, row2 = get_section(filename, "BIASSEC", firstext)
        col1 += 1

        col1 += margin_cols

        col2 -= 1
        row2 += overscanrows

        for i in range(firstext, lastext):
            # make data float32 for calculations
            im[i].data = im[i].data.astype("float32")

            Median.append([])  # first hdu will be 1
            for row in range(row1, row2 + 1):
                Median[i].append(
                    numpy.median(im[i].data[row : row + 1, col1 : col2 + 1])
                )

            if fit_order > 0:
                slope, xdata, yfit, resids, residspercent = _line_fit(
                    list(range(row1, row2 + 1)), Median[i], fit_order
                )
                yfit = yfit.astype("float32")
            else:
                yfit = numpy.array(Median)

            # correct data by subtracting row by row best fit
            for row in range(row1, row2 + 1):
                if fit_order > 0:
                    im[i].data[row : row + 1] -= yfit[row]  # wraps here
                else:
                    im[i].data[row : row + 1] -= Median[i][row]

            # convert back to uint16, clipping to zero
            # im[i].data=im[i].data.clip(min=0)
            # im[i].data=im[i].data.astype('uint16')

        # new due to lock
        history_string = "COLBIAS data was column overscan corrected"
        value = (
            time.strftime("%Y/%m/%d %H:%M:%S ", time.gmtime(time.time()))
            + history_string
        )
        if len(value) > 70:
            value = value[:70] + "\n" + value[70:]
        if len(value) > 141:
            value = value[:141] + "\n" + value[141:]
        if len(value) > 212:
            value = value[:212]
        im[0].header.add_history(value)

    return


# *********************************************************************************************
# FITS support methods
# *********************************************************************************************
def _get_data_roi(azcam_roi: list = []) -> list:
    """
    Converts an azcam ROI to a numpy ROI.

    Args:
        azcam_roi: an azcam ROI list, i.e. one-based, columns first:
        [first_col, last_col, first_row, last_row].
    Returns:
        the equivalent python ROI, zero-based, rows first: [first_row,last_row,first_col,last_col].
    """

    if azcam_roi == []:
        roi = azcam.db.imageroi if azcam.db.get("imageroi") else []
    else:
        roi = azcam_roi

    if isinstance(roi[0], list):
        roi = roi[0]  # use first roi

    first_row = int(roi[2]) - 1
    last_row = int(roi[3]) - 1 + 1  # slice
    first_col = int(roi[0]) - 1
    last_col = int(roi[1]) - 1 + 1  # slice

    return [first_row, last_row, first_col, last_col]


def _line_fit(xdata, ydata, order=1, fit_min=0, fit_max=-1):
    """
    Fit a line to Data, usings points fit_min through fit_max.

    Args:
        xdata and ydata must be same-sized arrays.
    Returns:
        list `[Status,[slope,Y-intercept],xdata,YFit,residuals,residuals_percent]`.
    """

    num_points = len(xdata)
    if len(ydata) != num_points:
        return "ERROR X and Y arrays are not the same size"

    if fit_max == -1:
        fit_max = num_points - 1

    if fit_min > fit_max:
        return "ERROR fit_max must be <= fit_min"

    # make least squares fit through [fit_min:fit_max] points
    xxdata = xdata[fit_min:fit_max]
    yydata = ydata[fit_min:fit_max]

    # generate line y values
    try:
        polycoeffs = poly.polyfit(
            xxdata, yydata, order
        )  # [slope,intercept] using just fit_min:fit_max
    except Exception:
        polycoeffs = poly.polyfit(
            xxdata, yydata, order
        )  # [slope,intercept] using just fit_min:fit_max
    yfit = poly.polyval(xdata, polycoeffs)  # fit for all data, not just xxdata

    # calculate residuals
    residuals = []
    residuals_percent = []
    for i in range(num_points):
        r1 = ydata[i] - yfit[i]  # difference
        residuals.append(r1)  # residuals
        r2 = 100.0 * r1 / yfit[i]  # residual %
        residuals_percent.append(r2)  # residuals %

    return [polycoeffs, xdata, yfit, residuals, residuals_percent]



"""
Contains the ExposureMag class.
"""

import os
import time

import azcam
from azcam.tools.exposure import Exposure

from .receive_data import ReceiveData


class ExposureMag(Exposure):
    """
    Defines the exposure class for Magellan controllers which makes an exposure.
    """

    def __init__(self, tool_id="exposure", description=None):
        super().__init__(tool_id, description)

        self.receive_data = ReceiveData(self)
        self.exp_start = 0
        self.curr_delay = 0

    def integrate(self):
        """
        Integration.
        """

        # start integration
        self.exposure_flag = self.exposureflags["EXPOSING"]
        imagetype = self.image_type.lower()

        # do this as some DSP code not work right
        try:
            shutterstate = self.shutter_dict[self.image_type.lower()]
        except KeyError:
            shutterstate = 1  # other types are comps, so open shutter

        if shutterstate:
            azcam.db.tools["controller"].set_shutter(1)

        # start exposure
        if imagetype != "zero":
            azcam.log("Integration started")

        self.exp_start = time.time()
        azcam.db.tools["controller"].start_exposure()
        self.dark_time_start = time.time()

        # Mag controller pause/resume not supported but abort is
        if self.exposure_time >= 1.0:
            while self.exposure_time_remaining > 0.15:
                if self.exposure_flag == self.exposureflags["ABORT"]:
                    if self.is_exposure_sequence:
                        azcam.log("Stopping exposure sequence")
                        self.is_exposure_sequence = 0
                        self.exposure_sequence_number = 1
                        self.exposure_flag = self.exposureflags["EXPOSING"]
                    else:
                        azcam.db.tools["controller"].exposure_abort()
                    break
                time.sleep(0.1)
                self.exposure_time_remaining = self.exposure_time_remaining - 0.1
        else:
            time.sleep(self.exposure_time)

        azcam.db.tools["controller"].set_shutter(0)  # until shutter issue is solved

        # exposure finished
        if imagetype == "zero":
            self.exposure_time = self.exposure_time_saved
        self.exposure_time_remaining = 0
        if self.exposure_flag == self.exposureflags["ABORT"]:
            azcam.log("Integration aborted")
        else:
            self.exposure_flag = self.exposureflags["READ"]

        if shutterstate:
            azcam.db.tools["controller"].set_shutter(0)

        if self.image_type.lower() != "zero":
            azcam.log("integration finished", level=2)

        time.sleep(0.04)  # some bug on controller

        return

    def get_exposuretime_remaining(self):
        """
        Return remaining exposure time (in seconds).
        """

        return self.exposure_time_remaining

    def readout(self):
        """
        Exposure readout.
        """

        self.exposure_flag = self.exposureflags["READ"]

        imagetype = self.image_type.lower()

        if imagetype == "ramp":
            azcam.db.tools["controller"].set_shutter(1)

        # start readout
        azcam.log("Readout started")

        reply = azcam.db.tools["controller"].start_readout()

        # Wait for end of readout
        t_wait = (self.image.focalplane.numpix_image * 2) / 1000000 + 0.1

        t_wait = self.curr_delay
        time.sleep(t_wait)

        self.pixels_remaining = 0
        if self.exposure_flag != self.exposureflags["ABORT"]:
            self.exposure_flag = self.exposureflags["NONE"]

        # transfer image data already read from controller
        try:
            reply = self.receive_data.receive_image_data(
                self.image.focalplane.numpix_image * 2
            )
        except azcam.exceptions.AzcamError:
            self.exposure_flag = self.exposureflags["ABORT"]

        self.image.valid = 1  # new

        if imagetype == "ramp":
            azcam.db.tools["controller"].set_shutter(0)

        if self.exposure_flag == self.exposureflags["ABORT"]:
            azcam.log("Readout aborted")
            return ["ABORTED", "Readout aborted"]
        else:
            azcam.log("Readout finished", level=2)
            self.exposure_flag = self.exposureflags["NONE"]
            return

    def end(self):
        """
        Completes an exposure by writing file and displaying image.
        """

        self.exposure_flag = self.exposureflags["WRITING"]

        if self.send_image:
            local_file = self.temp_image_file + "." + self.get_extname(self.filetype)
            try:
                os.remove(local_file)
            except FileNotFoundError:
                pass
        else:
            local_file = self.get_filename()

        # wait for image data to be received
        loop = 0
        while not self.image.valid and loop < 100:
            loop += 1
            time.sleep(0.050)
            if loop >= 100:
                azcam.log("ERROR image data not received in time")

        # update controller header with keywords which might have changed
        et = float(int(self.exposure_time_actual * 1000.0) / 1000.0)
        dt = float(int(self.dark_time * 1000.0) / 1000.0)
        azcam.db.headers["exposure"].set_keyword(
            "EXPTIME", et, "Exposure time (seconds)", "float"
        )
        azcam.db.headers["exposure"].set_keyword(
            "DARKTIME", dt, "Dark time (seconds)", "float"
        )

        # write file(s) to disk
        if self.save_file:
            azcam.log("Writing %s" % local_file)

            # write the file to disk
            self.image.overwrite = self.overwrite
            self.image.test_image = self.test_image
            reply = self.image.write_file(local_file, self.filetype)
            azcam.log("Writing finished", level=2)

            # set flag that image now written to disk
            self.image.is_written = 1

            if self.send_image:
                azcam.log("Sending image")
                self.sendimage.send_image(local_file, self.get_filename())

        # image data and file are now ready
        self.image.toggle = 1

        # display image
        if self.display_image:
            try:
                azcam.log("Displaying image")
                azcam.db.tools["display"].display(self.image)
            except Exception:
                pass

        # increment file sequence number if image was written
        if self.save_file:
            self.increment_filenumber()

        self.exposure_flag = self.exposureflags["NONE"]

        return



"""
DASH web server for azcam
"""

import threading
import logging

from dash import Dash, html, dcc, callback, Input, Output, State, ctx
import dash_bootstrap_components as dbc
import dash_daq as daq

import azcam
from azcam.web.exposure.exposure_card import exposure_card
from azcam.web.exposure.status_card import status_card
from azcam.web.exposure.filename_card import filename_card
from azcam.web.exposure.detector_card import detector_card
from azcam.web.exposure.options_card import options_card
from azcam.web.exposure.tabs import tabs_buttons, tabs


class ExposureWeb(object):
    """
    Web server using Dash (plotly).
    """

    def __init__(self) -> None:

        azcam.db.expserver = self

        self.app = Dash(
            __name__,
            external_stylesheets=[dbc.themes.BOOTSTRAP],
            suppress_callback_exceptions=False,
            requests_pathname_prefix="/exposure/",
        )
        logging.getLogger("werkzeug").setLevel(logging.CRITICAL)  # stop messages

        self.exposure_card = exposure_card()
        self.status_card = status_card()
        self.filename_card = filename_card()
        self.detector_card = detector_card()
        self.options_card = options_card()

        self.tabs_buttons = tabs_buttons()
        self.tabs = tabs(self)

        # app layout
        self.app.layout = html.Div(
            [
                self.tabs_buttons,
                self.tabs,
                html.Div(id="hidden_div", hidden=True),
                dcc.Interval("expstatus_interval", interval=1_000, n_intervals=0),
            ]
        )

        return

    def set_message(self, message: str = "") -> None:
        """
        Set the browser message.
        """

        azcam.db.tools["exposure"].message = repr(message)
        return
